@@ -20,12 +20,13 @@
    Boston, MA 02111-1307, USA.
 */
 
+#include <strings.h>
 
 #include <gtk/gtk.h>
 #include <gdk/gdkkeysyms.h>
 
 #include "document/htmlfocusiterator.h"
-#include "document/htmlparser.h"
+#include "document/htmldocument.h"
 #include "dom/core/dom-node.h"
 #include "dom/core/dom-element.h"
 #include "dom/views/dom-abstractview.h"
@@ -87,13 +88,18 @@ HTML_ACCESSIBLE_FACTORY (HTML_TYPE_BOX_TABLE_ACCESSIBLE, html_box_table_accessib
 
 static GQuark quark_moving_focus_out = 0;
 static GQuark quark_cursor_position = 0;
+static GQuark quark_cursor_end_of_line = 0;
 static GQuark quark_selection_bound = 0;
 static GQuark quark_layout = 0;
 static GQuark quark_cursor_visible = 0;
 static GQuark quark_blink_timeout = 0;
 static GQuark quark_button = 0;
+static GQuark quark_virtual_cursor_x = 0;
+static GQuark quark_virtual_cursor_y = 0;
 static gboolean cursor_showing = FALSE;
 
+HtmlBoxText* _html_view_get_cursor_box_text (HtmlView *view, gint *offset);
+
 static HtmlBox*
 find_last_child (HtmlBox *box)
 {
@@ -213,7 +219,7 @@ find_next_box_text (HtmlBox *box)
 }
 
 static HtmlBoxText*
-html_view_get_box_text_for_offset (HtmlView *view, gint *offset, gboolean forward)
+html_view_get_box_text_for_offset (HtmlView *view, gint *offset, gboolean end_of_line)
 {
 	HtmlBox *box;
 	HtmlBoxText *last_text;
@@ -232,7 +238,7 @@ html_view_get_box_text_for_offset (HtmlView *view, gint *offset, gboolean forwar
 			if (len > 0) {
 				if (len > *offset) {
 					return text;
-				} else if (!forward && *offset == len) {
+				} else if (end_of_line && *offset == len) {
 					return text;	
 				}
 				*offset -= len;
@@ -295,6 +301,75 @@ html_view_get_offset_for_box_text (HtmlView *view, HtmlBoxText *text, gint *offs
 	
 }
 
+static gint
+html_view_get_virtual_cursor_x (HtmlView *view)
+{
+	gint val;
+
+	if (quark_virtual_cursor_x == 0)
+		return -1;
+	val = GPOINTER_TO_INT (g_object_get_qdata (G_OBJECT (view),
+						   quark_virtual_cursor_x));
+	return val;
+}
+
+static void
+html_view_set_virtual_cursor_x (HtmlView *view, gint x)
+{
+	if (quark_virtual_cursor_x == 0)
+		quark_virtual_cursor_x = g_quark_from_static_string ("html-view-virtual-cursor-x");
+
+	g_object_set_qdata (G_OBJECT (view), 
+			    quark_virtual_cursor_x,
+			    GINT_TO_POINTER (x));
+}
+
+static gint
+html_view_get_virtual_cursor_y (HtmlView *view)
+{
+	gint val;
+
+	if (quark_virtual_cursor_y == 0)
+		return -1;
+	val = GPOINTER_TO_INT (g_object_get_qdata (G_OBJECT (view),
+						   quark_virtual_cursor_y));
+	return val;
+}
+
+static void
+html_view_set_virtual_cursor_y (HtmlView *view, gint y)
+{
+	if (quark_virtual_cursor_y == 0)
+		quark_virtual_cursor_y = g_quark_from_static_string ("html-view-virtual-cursor-y");
+
+	g_object_set_qdata (G_OBJECT (view), 
+			    quark_virtual_cursor_y,
+			    GINT_TO_POINTER (y));
+}
+
+static gint
+html_view_get_cursor_end_of_line (HtmlView *view)
+{
+	gint val;
+
+	if (quark_cursor_end_of_line == 0)
+		return 0;
+	val = GPOINTER_TO_INT (g_object_get_qdata (G_OBJECT (view),
+						   quark_cursor_end_of_line));
+	return val;
+}
+
+static void
+html_view_set_cursor_end_of_line (HtmlView *view, gint val)
+{
+	if (quark_cursor_end_of_line == 0)
+		quark_cursor_end_of_line = g_quark_from_static_string ("html-view-cursor-end-of-line");
+
+	g_object_set_qdata (G_OBJECT (view), 
+			    quark_cursor_end_of_line,
+			    GINT_TO_POINTER (val));
+}
+
 static gint
 html_view_get_cursor_position (HtmlView *view)
 {
@@ -319,7 +394,10 @@ html_view_get_cursor_position (HtmlView *view)
 			val1 = view->sel_end_index;
 		}
 		if (html_view_get_offset_for_box_text (view, text, &val)) {
-			val += val1;
+			gchar *str;
+
+			str = html_box_text_get_text (text, NULL);
+			val += g_utf8_pointer_to_offset (str, str + val1);
 			return val;
 		} else {
 			g_warning ("No offset for cursor position");
@@ -338,9 +416,29 @@ html_view_set_cursor_position (HtmlView *view, gint cursor_position)
 	if (quark_cursor_position == 0)
 		quark_cursor_position = g_quark_from_static_string ("html-view-cursor-position");
 
+	html_view_set_virtual_cursor_x (view, -1);
+	html_view_set_virtual_cursor_y (view, -1);
 	g_object_set_qdata (G_OBJECT (view), 
 			    quark_cursor_position,
 			    GINT_TO_POINTER (cursor_position));
+#ifdef ENABLE_ACCESSIBILITY
+	{
+	AtkObject *obj;
+	HtmlBoxText *box_text;
+	gint offset;
+
+	box_text = _html_view_get_cursor_box_text (view, NULL);	
+	if (box_text) {
+		obj = atk_gobject_accessible_for_object (G_OBJECT (box_text));
+		if (!ATK_IS_NO_OP_OBJECT (obj)) {
+			/* Accessibility is enabled */
+			g_return_if_fail (ATK_IS_TEXT (obj));
+			offset = atk_text_get_caret_offset (ATK_TEXT (obj));
+			g_signal_emit_by_name (obj, "text-caret-moved", offset);
+		}
+	}
+	}
+#endif /* ENABLE_ACCESSIBILITY */
 }
 
 static gint
@@ -367,7 +465,10 @@ html_view_get_selection_bound (HtmlView *view)
 			val1 = view->sel_start_index;
 		}
 		if (html_view_get_offset_for_box_text (view, text, &val)) {
-			val += val1;
+			gchar *str;
+
+			str = html_box_text_get_text (text, NULL);
+			val += g_utf8_pointer_to_offset (str, str + val1);
 			return val;
 		} else {
 			g_warning ("No offset for selection bound");
@@ -623,46 +724,91 @@ html_view_draw_insertion_cursor (GtkWidget        *widget,
 	}
 }
 
+
+static void
+html_view_get_box_text_location (HtmlView *view, HtmlBoxText *text, gint *offset,  GdkRectangle *pos)
+{
+	gint x, y;
+	gchar *str;
+	gchar *new_str;
+	HtmlBox *box;
+
+	str = html_box_text_get_text (text, NULL);
+	new_str = g_utf8_offset_to_pointer (str, *offset);
+	html_box_text_get_character_extents (text, (gint) (new_str - str), pos);
+	box = HTML_BOX (text);
+	x = html_box_get_absolute_x (box);
+	y = html_box_get_absolute_y (box);
+	pos->x += x - box->x;
+	pos->y += y - box->y;
+	pos->width = 0;
+
+	return;
+}
+
+HtmlBoxText*
+_html_view_get_cursor_box_text (HtmlView *view, gint *offset)
+{
+	HtmlBoxText *text;
+	gint cursor_position;
+	gint end_of_line;
+
+        cursor_position = html_view_get_cursor_position (view);
+
+	end_of_line = html_view_get_cursor_end_of_line (view);
+	text = html_view_get_box_text_for_offset (view, &cursor_position, end_of_line);
+	if (offset) {
+		*offset = cursor_position;
+	}
+		
+	if (text == NULL) {
+		g_assert (cursor_position <= 0);
+		return NULL;
+	}
+
+	return text;
+}
+
+static HtmlBox*
+html_view_get_cursor_location (HtmlView *view, GdkRectangle *pos)
+{
+	gint cursor_position;
+	HtmlBoxText *box_text;
+
+        box_text = _html_view_get_cursor_box_text (view, &cursor_position);
+	if (box_text) {
+		html_view_get_box_text_location (view, box_text, &cursor_position, pos);
+		return HTML_BOX (box_text);
+	} else {
+		return NULL;
+	}
+}
+
 static void 
 html_view_draw_cursor (HtmlView *view)
 {
-	gint cursor_position;
 	GdkRectangle rect;
-	gint x, y;
-	HtmlBoxText *text;
-	HtmlBox *box;
 	HtmlGdkPainter *gdk_painter;
 	gint direction;
 	GtkTextDirection text_direction;
+	HtmlBox *box;
 
 	if (!cursor_showing)
 		return;
 
-        cursor_position = html_view_get_cursor_position (view);
-	text = html_view_get_box_text_for_offset (view, &cursor_position, TRUE);
-	if (text == NULL) {
-		/*
-		 * Text may be deleted so do not attempt to display cursor
-		 */
-		g_assert (cursor_position <= 0);
+	box = html_view_get_cursor_location (view, &rect);
+	if (box == NULL)
 		return;
-	}
-	html_box_text_get_character_extents (text, cursor_position, &rect);
-	box = HTML_BOX (text);
-	x = html_box_get_absolute_x (box);
-	y = html_box_get_absolute_y (box);
+
 	direction = html_box_get_bidi_level (box);
 	text_direction = (direction == HTML_DIRECTION_RTL) ? GTK_TEXT_DIR_RTL
 							   : GTK_TEXT_DIR_LTR;
-	rect.x += x - box->x;
-	rect.y += y - box->y;
-	rect.width = 0;
 	gdk_painter = HTML_GDK_PAINTER (view->painter); 
 	html_view_draw_insertion_cursor (GTK_WIDGET (view),
 					 gdk_painter->window,
 					 gdk_painter->gc,
 					 &rect,
-					 GTK_TEXT_DIR_LTR,
+					 text_direction,
 					 FALSE);
 }
 
@@ -808,12 +954,70 @@ html_view_pend_cursor_blink (HtmlView *view)
  	}
 }
 
+static gboolean
+is_box_in_paragraph (HtmlBox *box)
+{
+	while (box && !HTML_IS_BOX_BLOCK (box)) {
+		box = box->parent;
+	}
+	if (!box) {
+		return FALSE;
+	}
+	if (box->dom_node && strcmp ((char*)box->dom_node->xmlnode->name, "p") == 0) {
+		return TRUE;
+	}
+	return FALSE;
+}
+
+static gboolean
+is_offset_in_paragraph (HtmlView *view, gint offset)
+{
+	HtmlBoxText *text;
+	HtmlBox *box;
+	gint tmp_offset;
+
+	tmp_offset = offset;
+	text = html_view_get_box_text_for_offset (view, &tmp_offset, TRUE);
+	if (!text)
+		return FALSE;
+	box = HTML_BOX (text);
+	return is_box_in_paragraph (box);
+}
+
+static gboolean
+is_at_line_boundary (HtmlView *view, gint offset)
+{
+	HtmlBoxText *next;
+	HtmlBoxText *prev;
+	HtmlBox *next_box;
+	HtmlBox *prev_box;
+	gint tmp_offset;
+
+	tmp_offset = offset;
+	next = html_view_get_box_text_for_offset (view, &tmp_offset, FALSE);
+	if (!next || tmp_offset > 0)
+		return FALSE;
+	next_box = HTML_BOX (next);
+	prev = find_previous_box_text (next_box);
+	if (!prev)
+		return FALSE;
+
+	prev_box = HTML_BOX (prev);
+	if (html_box_get_absolute_y (next_box) == html_box_get_absolute_y (prev_box))
+		return FALSE;
+	else
+		return TRUE;
+}
+
 static gint
 html_view_move_visually (HtmlView *view, gint start, gint count)
 {
 	PangoLayout *layout;
 	const gchar* text;
 	gint index;
+	gint offset;
+	gboolean forward;
+	gint end_of_line;
 
 	html_view_setup_layout (view);
 	layout = html_view_get_layout (view);
@@ -822,30 +1026,68 @@ html_view_move_visually (HtmlView *view, gint start, gint count)
  
 	index = g_utf8_offset_to_pointer (text, start) - text;
 
-	while (count != 0) {
-		int new_index, new_trailing;
-		gboolean strong;
-
-		strong = TRUE;
-
-		if (count > 0) {
-			pango_layout_move_cursor_visually (layout, strong, index, 0, 1, &new_index, &new_trailing);
+	forward = count >= 0;
+	end_of_line = html_view_get_cursor_end_of_line (view);
+	if (forward) {
+		if (end_of_line == 1 &&
+		    is_at_line_boundary (view, start)) {
 			count--;
-		} else {
-			pango_layout_move_cursor_visually (layout, strong, index, 0, -1, &new_index, &new_trailing);
-			count++;
+			offset = start;
 		}
+	} else {
+		if (end_of_line == 0 &&
+		    is_at_line_boundary (view, start) &&
+		    !is_offset_in_paragraph (view, start)) {
+			count++;
+			offset = start;
+		}
+	}
 
-		if (new_index < 0 || new_index == G_MAXINT)
-			break;
+	if (count != 0) {
+		while (count != 0) {
+			int new_index, new_trailing;
+			gboolean strong;
 
-		index = new_index;
+			strong = TRUE;
 
-		while (new_trailing--)
-			index = g_utf8_next_char (text + new_index) - text;
+			if (count > 0) {
+				pango_layout_move_cursor_visually (layout, strong, index, 0, 1, &new_index, &new_trailing);
+				count--;
+			} else {
+				pango_layout_move_cursor_visually (layout, strong, index, 0, -1, &new_index, &new_trailing);
+				count++;
+			}
+
+			if (new_index < 0 || new_index == G_MAXINT)
+				break;
+
+			index = new_index;
+
+			while (new_trailing--)
+				index = g_utf8_next_char (text + new_index) - text;
+		}
+		offset = g_utf8_pointer_to_offset (text, text + index);
+	}
+	if (offset == start) {
+		if (forward)
+			html_view_set_cursor_end_of_line (view, 0);
+		else {
+			html_view_set_cursor_end_of_line (view, 1);
+		}
+	} else {
+		if (forward) {
+			if (is_at_line_boundary (view, offset) &&
+			    is_offset_in_paragraph (view, offset)) {
+				html_view_set_cursor_end_of_line (view, 0);
+			} else {
+				html_view_set_cursor_end_of_line (view, 1);
+			}
+		} else {
+			html_view_set_cursor_end_of_line (view, 0);
+		}
 	}
  
-	return g_utf8_pointer_to_offset (text, text + index);
+	return offset;
 }
 
 static gint
@@ -866,7 +1108,7 @@ html_view_move_backward_word (HtmlView *view, gint start)
 
 	new_pos--;
 	box_offset = new_pos;
-	text = html_view_get_box_text_for_offset (view, &box_offset, TRUE);
+	text = html_view_get_box_text_for_offset (view, &box_offset, FALSE);
 	box_offset = new_pos - box_offset;
 	/* Find the previous word beginning or start of current box*/
 	while (new_pos > 0 && (!log_attrs[new_pos].is_word_start &&
@@ -874,6 +1116,7 @@ html_view_move_backward_word (HtmlView *view, gint start)
 		new_pos--;
 
 	g_free (log_attrs);
+	html_view_set_cursor_end_of_line (view, 0);
 	return new_pos;
 }
 
@@ -900,7 +1143,7 @@ html_view_move_forward_word (HtmlView *view, gint start)
 	/* Find the next word end or end of current box*/
 	new_pos++;
 	box_offset = new_pos;
-	text = html_view_get_box_text_for_offset (view, &box_offset, TRUE);
+	text = html_view_get_box_text_for_offset (view, &box_offset, FALSE);
 	char_text = html_box_text_get_text (text, &len);
 	len = g_utf8_strlen (char_text, len);
 	len += new_pos - box_offset;
@@ -909,9 +1152,55 @@ html_view_move_forward_word (HtmlView *view, gint start)
 		new_pos++;
 
 	g_free (log_attrs);
+	html_view_set_cursor_end_of_line (view, 1);
 	return new_pos;
 }
 
+static void
+html_view_get_virtual_cursor_pos (HtmlView *view, gint *x, gint *y)
+{
+	gint virtual_x, virtual_y;
+	gint cursor_x, cursor_y;
+	GdkRectangle pos;
+
+	virtual_x = html_view_get_virtual_cursor_x (view);
+	virtual_y = html_view_get_virtual_cursor_y (view);
+
+	if ((virtual_x == -1) ||
+	    (virtual_y == -1)) {
+        	html_view_get_cursor_location (view, &pos);
+	}
+
+	if (x) {   
+		if (virtual_x != -1) {
+			*x = virtual_x;
+		} else {
+			*x = pos.x;
+		}
+	}
+			
+	if (y) {   
+		if (virtual_y != -1) {
+			*y = virtual_y;
+		} else {
+			*y = pos.y + pos.height/2;
+		}
+	}
+}
+
+static void
+html_view_set_virtual_cursor_pos (HtmlView *view, gint x, gint y)
+{
+	GdkRectangle pos;
+
+	if (x == -1 || y == -1) {
+		html_view_get_cursor_location (view, &pos);
+	}
+
+	html_view_set_virtual_cursor_x (view, (x == -1) ? pos.x : x);
+	html_view_set_virtual_cursor_y (view, (y == -1) ? pos.y  + pos.height / 2 : y);
+}
+
 static gboolean
 is_moving_focus_out (HtmlView *view)
 {
@@ -959,52 +1248,230 @@ set_adjustment_clamped (GtkAdjustment *adj, gdouble val)
     return FALSE;
 }
 
+static HtmlBoxText*
+find_box_text_for_x_pos (HtmlView *view, HtmlBoxText* text, gboolean forward, gint x_pos)
+{
+	HtmlBoxText *new_text;
+	HtmlBox *new_box;
+	HtmlBoxText *prev_text;
+	HtmlBox *prev_box;
+	gint y;
+	gint x;
+	gint new_x;
+	gint new_y;
+
+	prev_box = HTML_BOX (text);
+	x = html_box_get_absolute_x (prev_box);
+	if (forward) {
+		if (x + prev_box->width > x_pos) {
+			return text;
+		}
+	} else {
+		if (x <= x_pos) {
+			return text;
+		}
+	}
+	y = html_box_get_absolute_y (prev_box);
+	prev_text = text;
+	while (TRUE) {
+		if (forward) {
+			new_text = find_next_box_text (prev_box);
+			if (!new_text)
+				return prev_text;
+			new_box = HTML_BOX (new_text);
+			new_y = html_box_get_absolute_y (new_box);
+			if (new_y > y) {
+				return prev_text;
+			}
+			new_x = html_box_get_absolute_x (new_box);
+			if (new_x + new_box->width > x_pos)
+				return new_text;
+			prev_text = new_text;
+			prev_box = new_box;
+		} else {
+			new_text = find_previous_box_text (prev_box);
+			if (!new_text)
+				return prev_text;
+			new_box = HTML_BOX (new_text);
+			new_y = html_box_get_absolute_y (new_box);
+			if (new_y < y) {
+				return prev_text;
+			}
+			new_x = html_box_get_absolute_x (new_box);
+			if (new_x <= x_pos)
+				return new_text;
+			prev_text = new_text;
+			prev_box = new_box;
+		}
+	}
+}
+
+static gint 
+set_offset_for_box_text (HtmlView *view, HtmlBoxText *text, gint x_pos)
+{
+	gint x;
+	HtmlBox *box;
+	gint tmp_offset;
+	gint offset;
+	char *char_text;
+
+	box = HTML_BOX (text);
+	x = html_box_get_absolute_x (box);
+
+	if (x + box->width <= x_pos) {
+		gint len;
+
+		char_text = html_box_text_get_text (text, &len);
+		tmp_offset = g_utf8_strlen (char_text, len);
+		if (is_box_in_paragraph (box)) {
+			tmp_offset--;
+			html_view_set_cursor_end_of_line (view, 0);
+		} else {
+			html_view_set_cursor_end_of_line (view, 1);
+		}
+	} else {
+		if (x_pos > x)
+			x = x_pos - x;
+		else 
+			x = 0;
+		tmp_offset = html_box_text_get_index (text, x);
+		char_text = html_box_text_get_text (text, NULL);
+		tmp_offset = g_utf8_pointer_to_offset (char_text, char_text + tmp_offset);
+		html_view_set_cursor_end_of_line (view, 0);
+	}
+	html_view_get_offset_for_box_text (view, text, &offset);
+	offset += tmp_offset;
+	return offset;
+} 
+
+static void
+move_cursor (HtmlView *view, HtmlBox *new_cursor_box, gint offset, gboolean extend_selection)
+{
+	gint selection_bound;
+	gint cursor_position;
+
+	cursor_position = html_view_get_cursor_position (view);
+	selection_bound = html_view_get_selection_bound (view);
+	if (extend_selection && cursor_position != offset) {
+		HtmlBoxText *start_text;
+		HtmlBoxText *end_text;
+		HtmlBox *start_box;
+		HtmlBox *end_box;
+		gchar *start_str;
+		gchar *end_str;
+		gint start_offset;
+		gint end_offset;
+		gint extent;
+		gboolean forward;
+
+		extent = offset - selection_bound;
+		forward = extent >= 0;
+		if (forward) {
+			start_offset = selection_bound;
+			end_offset = offset;
+		} else {
+			start_offset = offset;
+			end_offset = selection_bound;
+			extent = -extent;
+		}
+		start_text = html_view_get_box_text_for_offset (view, &start_offset, FALSE);
+		end_text = html_view_get_box_text_for_offset (view, &end_offset, TRUE);
+		html_view_set_cursor_position (view, offset);
+		start_box = HTML_BOX (start_text);
+		end_box = HTML_BOX (end_text);
+		html_selection_extend (view, start_box, start_offset, extent);
+
+		view->sel_start = start_box;
+		start_str = html_box_text_get_text (start_text, NULL);
+		view->sel_start_index = 
+			g_utf8_offset_to_pointer (start_str, start_offset) 
+				- start_str;
+		view->sel_end = end_box;
+		end_str = html_box_text_get_text (end_text, NULL);
+		view->sel_end_index = 
+			g_utf8_offset_to_pointer (end_str, end_offset) 
+				- end_str;
+		view->sel_backwards = !forward;	
+	} else {
+		HtmlBoxText *text;
+
+		html_view_set_cursor_position (view, offset);
+		html_view_set_selection_bound (view, offset);
+
+		if (new_cursor_box == NULL) {
+			text = html_view_get_box_text_for_offset (view, &offset, html_view_get_cursor_end_of_line (view) != 0);
+			new_cursor_box = HTML_BOX (text);
+		}
+		if (DOM_IS_ELEMENT (new_cursor_box->parent->dom_node)) {
+			DomElement *element;
+
+			element = DOM_ELEMENT (new_cursor_box->parent->dom_node);
+			if (dom_element_is_focusable (element) &&
+			    element != view->document->focus_element) {
+                		html_document_update_focus_element (view->document, element);
+                		html_view_focus_element (view);
+			}
+		}
+	}
+
+}
+
+/*
+ * Move cursor to x_pos in current or previous line.
+ */
 static HtmlBox*
-html_view_move_cursor_by_line (HtmlView *view, gint count, gint *pos)
+html_view_move_cursor_by_line (HtmlView *view, gint count, gint x_pos, gint *offset)
 {
 	HtmlBoxText *text;
 	HtmlBox *box;
 	HtmlBox *last_box;
-	gint x, y, height;
-	gint old_x, old_y, old_height;
+	gint tmp_offset;
+	gint y, height;
+	gint old_y, old_height;
 	GtkAdjustment *adj;
-	gint tmp_pos;
+	gint end_of_line;
 
-	tmp_pos = *pos;
-	text = html_view_get_box_text_for_offset (view, &tmp_pos, TRUE);
+	tmp_offset = *offset;
+	end_of_line = html_view_get_cursor_end_of_line (view);
+	text = html_view_get_box_text_for_offset (view, &tmp_offset, end_of_line != 0);
 	box = HTML_BOX (text);
-	old_x = html_box_get_absolute_x (box);
 	old_y = html_box_get_absolute_y (box);
-	y = old_y;
 	old_height = box->height;
-	while (count > 0) {
-		text = find_next_box_text (box);
-		if (!text)
-			return NULL;
-		box = HTML_BOX (text);
-		x = html_box_get_absolute_x (box);
-		y = html_box_get_absolute_y (box);
-		if (old_y + old_height <= y) {
-			count--;
-		}
-	}
-	while (count < 0) {
-		text = find_previous_box_text (box);
-		if (!text)
-			return NULL;
-		box = HTML_BOX (text);
-		x = html_box_get_absolute_x (box);
-		y = html_box_get_absolute_y (box);
-		height = box->height;
-		if (y + height <= old_y) {
-			count++;
-		}
-	}
-	html_view_get_offset_for_box_text (view, text, pos);
 	adj = GTK_LAYOUT (view)->vadjustment;
-	if (!(adj->value && adj->value + adj->page_size > y)) {
-		set_adjustment_clamped (adj, y - adj->page_size + box->height);
+	if (count > 0) {
+		while (count > 0) {
+			text = find_next_box_text (box);
+			if (!text) {
+				return NULL;
+			}
+			box = HTML_BOX (text);
+			y = html_box_get_absolute_y (box);
+			if (old_y + old_height <= y) {
+				text = find_box_text_for_x_pos (view, text, TRUE, x_pos);
+				break;
+			}
+		}
+		if (!(adj->value + adj->page_size > y))
+			set_adjustment_clamped (adj, y - adj->page_size + box->height);
+	} else if (count < 0) {
+		while (count < 0) {
+			text = find_previous_box_text (box);
+			if (!text) {
+				return NULL;
+			}
+			box = HTML_BOX (text);
+			y = html_box_get_absolute_y (box);
+			height = box->height;
+			if (y + height <= old_y ) {
+				text = find_box_text_for_x_pos (view, text, FALSE, x_pos);
+				break;
+			}
+		}
+		if (!(adj->value <= y))
+			set_adjustment_clamped (adj, y);
 	}
+	*offset = set_offset_for_box_text (view, text, x_pos);
+	box = HTML_BOX (text);
 	return box;
 }
 
@@ -1036,7 +1503,7 @@ get_end_text_offset (HtmlView *view, gint *pos)
 			len = g_utf8_strlen (char_text, len);
 			html_view_get_offset_for_box_text (view, text, &offset);
 			temp_offset = offset;
-			temp = html_view_get_box_text_for_offset (view, &temp_offset, TRUE);
+			temp = html_view_get_box_text_for_offset (view, &temp_offset, FALSE);
 			offset += len;
 			*pos = offset;
 			return HTML_BOX (text);
@@ -1066,7 +1533,7 @@ html_view_move_cursor_to_start (HtmlView *view, gint *pos)
 	GtkAdjustment *adj;
 
 	*pos = 0;
-	text = html_view_get_box_text_for_offset (view, pos, TRUE);
+	text = html_view_get_box_text_for_offset (view, pos, FALSE);
 	adj = GTK_LAYOUT (view)->vadjustment;
 	if (adj->value) {
 		set_adjustment_clamped (adj, 0);
@@ -1074,8 +1541,8 @@ html_view_move_cursor_to_start (HtmlView *view, gint *pos)
 	return HTML_BOX (text);
 }
 
-static HtmlBox*
-html_view_move_cursor_by_page (HtmlView *view, gint count, gint *pos)
+static void
+html_view_scroll_pages (HtmlView *view, gint count, gboolean extend_selection)
 {
 	HtmlBox *box;
 	HtmlBoxText *text;
@@ -1083,31 +1550,39 @@ html_view_move_cursor_by_page (HtmlView *view, gint count, gint *pos)
 	gint x, y, height;
 	gdouble new_value;
 	gdouble old_value;
+	gint cursor_x_pos;
+	gint cursor_y_pos;
+	gint offset;
+
+	offset = html_view_get_cursor_position (view);
 
 	adj = GTK_LAYOUT (view)->vadjustment;
 	if (count > 0 && adj->value >= (adj->upper - adj->page_size - 1e-12)) {
 		/* already at bottom, make sure we are at the end */
-		box = get_end_text_offset (view, pos);
-		return box;
+		box = get_end_text_offset (view, &offset);
+		move_cursor (view, NULL, offset, extend_selection);
+		return; 
 	} else if (count < 0 && adj->value <= (adj->lower + 1e-12)) {
 		/* already at top, make sure we are at ofset 0 */
-		*pos = 0;
-		text = html_view_get_box_text_for_offset (view, pos, TRUE);
-		return HTML_BOX (text);
+		offset = 0;
+		move_cursor (view, NULL, offset, extend_selection);
+		return;
 	}
+	html_view_get_virtual_cursor_pos (view, &cursor_x_pos, &cursor_y_pos);
 	new_value = adj->value;
 	old_value = adj->value;
 
 	new_value += count * adj->page_increment;
 	set_adjustment_clamped (adj, new_value);
+	cursor_y_pos += adj->value - old_value;
 	
-	text = html_view_get_box_text_for_offset (view, pos, TRUE);
+	text = html_view_get_box_text_for_offset (view, &offset, FALSE);
 	box = HTML_BOX (text);
 	y = html_box_get_absolute_y (box);
 	while (count > 0) {
 		text = find_next_box_text (box);
 		if (!text) {
-			return NULL;
+			return;
 		}
 		box = HTML_BOX (text);
 		y = html_box_get_absolute_y (box);
@@ -1116,22 +1591,28 @@ html_view_move_cursor_by_page (HtmlView *view, gint count, gint *pos)
 		}
 	}
 	while (count < 0) {
-		HtmlBox *previous_box;
+		HtmlBoxText *previous_text;
 
-		previous_box = box;
+		previous_text = text;
 		text = find_previous_box_text (box);
 		if (!text) {
-			return NULL;
+			count = 0;
+			text = previous_text;
+			box = HTML_BOX (text);
 		}
 		box = HTML_BOX (text);
 		y = html_box_get_absolute_y (box);
 		if (y < adj->value) {
 			count = 0;
-			box = previous_box;
+			text = previous_text;
+			box = HTML_BOX (text);
 		}
 	}
-	html_view_get_offset_for_box_text (view, box, pos);
-	return box;
+	text = find_box_text_for_x_pos (view, text, TRUE, cursor_x_pos);
+	offset = set_offset_for_box_text (view, text, cursor_x_pos);
+	move_cursor (view, HTML_BOX (text), offset, extend_selection);
+	html_view_set_virtual_cursor_pos (view, cursor_x_pos, cursor_y_pos);
+	return;
 }
 
 void
@@ -2089,9 +2570,10 @@ html_view_real_move_cursor (HtmlView *html_view, GtkMovementStep step, gint coun
 {
 	GtkAdjustment *vertical, *horizontal;
 	HtmlBox *new_cursor_box;
-	gint new_pos;
+	gint new_offset;
 	gint selection_bound;
 	gint cursor_position;
+	gint cursor_x_pos = 0;
 
 	vertical = GTK_LAYOUT (html_view)->vadjustment;
 	horizontal = GTK_LAYOUT (html_view)->hadjustment;
@@ -2122,9 +2604,16 @@ html_view_real_move_cursor (HtmlView *html_view, GtkMovementStep step, gint coun
 		return;
 	}
 
+	if (step == GTK_MOVEMENT_PAGES) {
+		html_view_scroll_pages (html_view, count, extend_selection);
+		html_view_check_cursor_blink (html_view);
+		html_view_pend_cursor_blink (html_view);
+		return;
+	}
+
 	cursor_position = html_view_get_cursor_position (html_view);
 	selection_bound = html_view_get_selection_bound (html_view);
-	new_pos = cursor_position;
+	new_offset = cursor_position;
 	new_cursor_box = NULL;
 	if (cursor_position != selection_bound && !extend_selection) {
 	/*
@@ -2138,17 +2627,17 @@ html_view_real_move_cursor (HtmlView *html_view, GtkMovementStep step, gint coun
 			gint bound_x = get_better_cursor_x (html_view, selection_bound);
 
 			if (count < 0)
-				new_pos = current_x < bound_x ? cursor_position : selection_bound;
+				new_offset = current_x < bound_x ? cursor_position : selection_bound;
 			else
-				new_pos = current_x > bound_x ? cursor_position : selection_bound;
+				new_offset = current_x > bound_x ? cursor_position : selection_bound;
 
 			break;
           		}
         	case GTK_MOVEMENT_WORDS:
 			if (count < 0)
-				new_pos = MIN (cursor_position, selection_bound);
+				new_offset = MIN (cursor_position, selection_bound);
 			else
-				new_pos = MAX (cursor_position, selection_bound);
+				new_offset = MAX (cursor_position, selection_bound);
 			break;
 		default:
 			break;
@@ -2157,29 +2646,27 @@ html_view_real_move_cursor (HtmlView *html_view, GtkMovementStep step, gint coun
 	} else {
 		switch (step) {
 		case GTK_MOVEMENT_VISUAL_POSITIONS:
-			new_pos = html_view_move_visually (html_view, new_pos, count);
+			new_offset = html_view_move_visually (html_view, new_offset, count);
 			break;
         	case GTK_MOVEMENT_WORDS:
 			while (count > 0) {
-				new_pos = html_view_move_forward_word (html_view, new_pos);
+				new_offset = html_view_move_forward_word (html_view, new_offset);
 				count--;
 			}
 			while (count < 0) {
-				new_pos = html_view_move_backward_word (html_view, new_pos);
+				new_offset = html_view_move_backward_word (html_view, new_offset);
 				count++;
 			}
 			break;
 		case GTK_MOVEMENT_DISPLAY_LINES:
-			new_cursor_box = html_view_move_cursor_by_line (html_view, count, &new_pos);	
-			break;
-		case GTK_MOVEMENT_PAGES:
-			new_cursor_box = html_view_move_cursor_by_page (html_view, count, &new_pos);	
+			html_view_get_virtual_cursor_pos (html_view, &cursor_x_pos, NULL);
+			new_cursor_box = html_view_move_cursor_by_line (html_view, count, cursor_x_pos, &new_offset);	
 			break;
 		case GTK_MOVEMENT_BUFFER_ENDS:
 			if (count < 0) {
-				new_cursor_box = html_view_move_cursor_to_start (html_view, &new_pos);
+				new_cursor_box = html_view_move_cursor_to_start (html_view, &new_offset);
 			} else if (count > 0) {
-				new_cursor_box = html_view_move_cursor_to_end (html_view, &new_pos);
+				new_cursor_box = html_view_move_cursor_to_end (html_view, &new_offset);
 			}
 			break;
 		default:
@@ -2187,58 +2674,9 @@ html_view_real_move_cursor (HtmlView *html_view, GtkMovementStep step, gint coun
 		}
 	}
 
-	if (extend_selection && cursor_position != new_pos) {
-		HtmlBoxText *start_text;
-		HtmlBoxText *end_text;
-		HtmlBox *start_box;
-		HtmlBox *end_box;
-		gint start_pos;
-		gint end_pos;
-		gint extent;
-		gboolean forward;
-
-		extent = new_pos - selection_bound;
-		forward = extent >= 0;
-		if (forward) {
-			start_pos = selection_bound;
-			end_pos = new_pos;
-		} else {
-			start_pos = new_pos;
-			end_pos = selection_bound;
-			extent = -extent;
-		}
-		start_text = html_view_get_box_text_for_offset (html_view, &start_pos, TRUE);
-		end_text = html_view_get_box_text_for_offset (html_view, &end_pos, FALSE);
-		html_view_set_cursor_position (html_view, new_pos);
-		start_box = HTML_BOX (start_text);
-		end_box = HTML_BOX (end_text);
-		html_selection_extend (html_view, start_box, start_pos, extent);
-
-		html_view->sel_start = start_box;
-		html_view->sel_start_index = start_pos;
-		html_view->sel_end = end_box;
-		html_view->sel_end_index = end_pos;
-		html_view->sel_backwards = !forward;	
-	} else {
-		html_view_set_cursor_position (html_view, new_pos);
-		html_view_set_selection_bound (html_view, new_pos);
-
-		if (new_cursor_box == NULL) {
-			HtmlBoxText *text;
-
-			text = html_view_get_box_text_for_offset (html_view, &new_pos, TRUE);
-			new_cursor_box = HTML_BOX (text);
-		}
-		if (DOM_IS_ELEMENT (new_cursor_box->parent->dom_node)) {
-			DomElement *element;
-
-			element = DOM_ELEMENT (new_cursor_box->parent->dom_node);
-			if (dom_element_is_focusable (element) &&
-			    element != html_view->document->focus_element) {
-                		html_document_update_focus_element (html_view->document, element);
-                		html_view_focus_element (html_view);
-			}
-		}
+	move_cursor (html_view, new_cursor_box, new_offset, extend_selection);
+	if (step == GTK_MOVEMENT_DISPLAY_LINES) {
+		html_view_set_virtual_cursor_pos (html_view, cursor_x_pos, -1);
 	}
 	html_view_check_cursor_blink (html_view);
 	html_view_pend_cursor_blink (html_view);
@@ -2586,7 +3024,6 @@ html_view_init (HtmlView *view)
 	view->relayout_timeout_id = 0;
 	view->magnification = 1.0;
 	g_signal_connect (view, "toggle-cursor", G_CALLBACK (html_view_toggle_cursor), NULL);
-	html_view_set_cursor_position (view, 0);
 	html_view_set_layout (view, NULL);
 }
 
@@ -2773,15 +3210,29 @@ static void
 html_view_focus_element (HtmlView *view)
 {
 	HtmlBox *box;
+	gint offset;
 
 	if (view->document->focus_element) {
 		html_view_scroll_to_node (view, DOM_NODE (view->document->focus_element), HTML_VIEW_SCROLL_TO_BOTTOM);
 
 		box = html_view_find_layout_box (view, DOM_NODE (view->document->focus_element), FALSE);
-		if (box && HTML_IS_BOX_EMBEDDED (box)) 
+		if (box && HTML_IS_BOX_EMBEDDED (box)) { 
 			gtk_widget_child_focus (HTML_BOX_EMBEDDED (box)->widget, GTK_DIR_TAB_FORWARD);
-		else
+		}
+		else {
 			gtk_widget_grab_focus (GTK_WIDGET (view));
+			if (cursor_showing && 
+			    HTML_IS_BOX_TEXT (box->children)) {
+				HtmlBoxText *text;
+
+				text = HTML_BOX_TEXT (box->children);
+				if (html_view_get_offset_for_box_text (view, text, &offset)) {
+					move_cursor (view, text, offset, FALSE);
+					html_view_pend_cursor_blink (view);
+					html_view_check_cursor_blink (view);
+				}
+			}
+		}
 	} else {
 		/* No element focused to scroll to top */
 		GtkAdjustment *adj = GTK_LAYOUT (view)->vadjustment;
